#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2012 Liricus SRL <info@liricus.com.ar>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Fra


#===============================================================================
# DOCS
#===============================================================================

"""Defines the database model.

"""


#===============================================================================
# IMPORTS
#===============================================================================

import shutil
import hashlib
import random
import datetime
import os
import inspect
import time

from werkzeug import security
from flask_peewee import db
from flask_peewee import auth
from flaskext import uploads

from socialframe_server import settings
from socialframe_server import core
from socialframe_server import fileutils


#===============================================================================
# TABLES
#===============================================================================

_tables = []

def autocreatetable(cls):
    if not inspect.isclass(cls) or not issubclass(cls, db.Model):
        msg = "'{}' is not subclass of {}".format(
            cls.__name__, str(core.APP_DB.Model)
        )
        raise ValueError(msg)
    if cls not in _tables:
        _tables.append(cls)
    return cls


def autocreatetables():
    return tuple(_tables)


#===============================================================================
# META MODEL
#===============================================================================

class SocialFrameModel(core.APP_DB.Model):
   pass


#===============================================================================
# USER MODELS AND REGISTER FOR USE WITH AUTH
#===============================================================================

@autocreatetable
class User(SocialFrameModel, auth.BaseUser):
    """Defines table ``User``:

    +--------------+-------------------------------+
    | Name         | Type                          |
    +==============+===============================+
    | username     | peewee.TextField(unique=True) |
    +--------------+-------------------------------+
    | email        | peewee.TextField(unique=True) |
    +--------------+-------------------------------+
    | fname        | peewee.CharField()            |
    +--------------+-------------------------------+
    | lname        | peewee.CharField()            |
    +--------------+-------------------------------+
    | birth        | peewee.CharField()            |
    +--------------+-------------------------------+
    | sex          | peewee.Charfield()            |
    +--------------+-------------------------------+

    **Fields**
    :username:
        User email.
    :email:
        User email.
    :fname:
        First name of the user.
    :lname:
        Last name of the user.
    :birth:
        Date of birth of the user.
    :sex:
        Sex of the user.

    """
    SEX_MALE = "m"
    SEX_FEMALE = "f"

    STATUS_ACTIVE = "active"
    STATUS_SUSPENDED = "suspended"
    STATUS_INACTIVE = "inactive"

    is_superuser = db.BooleanField(default=False)
    is_enterprise = db.BooleanField(default=False)
    active = db.BooleanField(default=False)
    email = db.CharField(unique=True)
    password = db.CharField()
    fname = db.CharField()
    lname = db.CharField()
    birth = db.DateField()
    sex = db.CharField(null=True, max_length=1,
                       choices=[(SEX_FEMALE, SEX_FEMALE),
                                (SEX_MALE, SEX_MALE)])
    sub_path = db.TextField()
    last_login = db.DateTimeField(null=True)
    last_know_lang = db.CharField(default=settings.DEFAULT_LANG)

    status_datetime = db.DateTimeField()
    status = db.CharField(db_column="status",
                          default=STATUS_INACTIVE,
                          choices=[(STATUS_ACTIVE, STATUS_ACTIVE),
                                   (STATUS_INACTIVE, STATUS_INACTIVE),
                                   (STATUS_SUSPENDED, STATUS_SUSPENDED)])

    def to_dict(self):
        return {"email": self.email,
                 "fname": self.fname,
                 "lname": self.lname,
                 "sex": self.sex,
                 "birth": core.formatdate(self.birth),
                 "active": self.active,
                 "is_enterprise": self.is_enterprise,
                 "is_superuser": self.is_superuser,
                 "last_login": core.formatdatetime(self.last_login),
                 "last_know_lang": self.last_know_lang,
                 "status": self.status,
                 "status_datetime": core.formatdatetime(self.status_datetime)}

    def __unicode__(self):
        return self.email

    def save(self, force_insert=False):
        if not self.sub_path:
            self.sub_path = os.path.join(self.email[0],
                                         self.email[:2],
                                         self.email)
        if not self.status_datetime:
            self.status_datetime = datetime.datetime.now()
        if not core.RX_MAIL.match(self.email):
            msg = "'{}' has not email format".format(self.email)
            raise ValueError(msg)
        super(User, self).save(force_insert)
        if not Profile.select().where(Profile.user == self).count():
            profile = Profile()
            profile.user = self
            profile.save()
        for inv in BubbleInvitation.select().where(BubbleInvitation.email == self.email):
            inv.to_user = self
            inv.save()

    def delete_instance(self, recursive=False):
        if os.path.exists(self.path):
            shutil.rmtree(self.path)
        super(User, self).delete_instance(recursive)

    def set_status(self, status):
        self.status = status
        self.active = (status == self.STATUS_ACTIVE)
        self.status_datetime = datetime.datetime.now()

    @property
    def username(self):
        return self.email

    @property
    def path(self):
        return os.path.join(settings.IMAGES, self.sub_path)



#===============================================================================
# PROFILE
#===============================================================================

@autocreatetable
class Profile(SocialFrameModel):
    """Defines table ``Profile`` with all no relevant information of the user:

    +-----------------+-------------------------------------------+
    | Name            | Type                                      |
    +=================+===========================================+
    | user            | peewee.ForeignKeyField(User, unique=True) |
    +-----------------+-------------------------------------------+
    | operation_hash  | peewee.CharField()                        |
    +-----------------+-------------------------------------------+
    | hash_expiration | peewee.DateTimeField(null=True)           |
    +-----------------+-------------------------------------------+
    | operation       | peewee.CharField( *see below )            |
    +-----------------+-------------------------------------------+

    **Fields**

    :user:
        Foreign Key of table ``User``.

    :operation_hash:
        Hash identifier of the operation.

    :hash_expiration:
        Expiration date of said hash.veni aca

    :operation:
        Defined action to be done on a profile.

    """

    OPERATION_CREATE_USER = 0
    OPERATION_RESET_PASSWORD = 10
    OPERATION_DELETE_USER = 20
    OPERATION_REACTIVATE_USER = 30

    user = db.ForeignKeyField(User, unique=True)
    _operation_hash = db.CharField(db_column="operation_hash", null=True)
    _hash_datetime = db.DateTimeField(db_column="hash_datetime", null=True)
    _operation = db.IntegerField(db_column="operation", null=True,
        choices=[
            (OPERATION_CREATE_USER, OPERATION_CREATE_USER),
            (OPERATION_RESET_PASSWORD, OPERATION_RESET_PASSWORD),
            (OPERATION_DELETE_USER, OPERATION_DELETE_USER),
            (OPERATION_REACTIVATE_USER, OPERATION_REACTIVATE_USER)
        ]
    )

    @property
    def operation(self):
        return self._operation

    @operation.setter
    def operation(self, operation):
        self._operation = operation
        self._operation_hash = hashlib.sha256(str(random.random())).hexdigest()
        self._hash_datetime = datetime.datetime.now()

    @property
    def operation_hash(self):
        return self._operation_hash

    def validate_hash(self, operation, operation_hash):
        """Confirms the ``operation`` by its hash.

        **Parameters**
            :operation:
                Defined action to be done on a profile.

            :operation_hash:
                hash number of the given operation.

        """

        if operation == self.operation and self._hash_datetime:
            now = datetime.datetime.now()
            delta = datetime.timedelta(seconds=settings.HASH_EXPIRATION)
            return now < self._hash_datetime + delta \
                    and operation_hash == self._operation_hash
        return False


#===============================================================================
# BUBBLE TABLE
#===============================================================================

@autocreatetable
class Bubble(SocialFrameModel):
    """:

    +-----------------+-------------------------------------------+
    | Name            | Type                                      |
    +=================+===========================================+
    | user            | peewee.ForeignKeyField(core.APP_AUTH.User)|
    +-----------------+-------------------------------------------+
    | name            | peewee.Charfield()                        |
    +-----------------+-------------------------------------------+
    | listed          | peewee.BooleanField()                     |
    +-----------------+-------------------------------------------+
    | fsname          | peewee.CharField()                        |
    +-----------------+-------------------------------------------+
    | password        | peewee.Charfield(null=True)               |
    +-----------------+-------------------------------------------+
    | desc            | peewee.TextField(null=True)               |
    +-----------------+-------------------------------------------+
    | random          | peewee.Charfield()                        |
    +-----------------+-------------------------------------------+

    **Fields**

    :user:

    :name:
        Name of bubble.
    :listed:
        If it's listed or not.
    :fsname:
        File system name.
    :password:
        Bubble password
    :desc:
        Description of bubble.
    :random:


    """

    COMMUNITY_BUBBLE = "community"
    PRIVATE_BUBBLE = "private"
    ENTERPRISE_BUBBLE = "enterprise"

    user = db.ForeignKeyField(User)
    name = db.CharField()
    listed = db.BooleanField(default=False)
    fsname = db.CharField()
    community = db.BooleanField(default=False)
    password = db.CharField(null=True)
    desc = db.TextField(null=True)
    random = db.CharField()
    type = db.CharField(choices=[(COMMUNITY_BUBBLE, COMMUNITY_BUBBLE),
                                  (PRIVATE_BUBBLE, PRIVATE_BUBBLE),
                                  (ENTERPRISE_BUBBLE, ENTERPRISE_BUBBLE)])

    _buffer_time = 3600
    _buffer_size = settings.GIVE_ME_BUBBLES_MAX * 100

    _buffer_bubbles = None
    _buffer_resync_time = -1

    _buffer_e_bubbles = None
    _buffer_e_resync_time = -1

    @classmethod
    def random_enterprise_bubbles(cls):
        if not Bubble._buffer_e_bubbles or time.time() > Bubble._buffer_resync_time:
            query = Bubble.select()
            query = query.where(Bubble.listed == True)
            query = query.where(Bubble.password >> None)
            query = query.where(Bubble.type == Bubble.ENTERPRISE_BUBBLE)
            if random.randint(0, 1):
                query = query.order_by(Bubble.random.asc())
            else:
                query = query.order_by(Bubble.random.desc())
            query = query.limit(Bubble._buffer_size)
            Bubble._buffer_e_bubbles = tuple([bubble.id for bubble in query])
            Bubble._buffer_e_resync_time = time.time() + Bubble._buffer_time
        idfilter = set()
        if len(Bubble._buffer_e_bubbles) <= settings.GIVE_ME_BUBBLES_MAX:
            idfilter = Bubble._buffer_e_bubbles
        else:
            while len(idfilter) < settings.GIVE_ME_BUBBLES_MAX:
                idfilter.add(random.choice(Bubble._buffer_e_bubbles))
        query = Bubble.select().limit(settings.GIVE_ME_BUBBLES_MAX)
        if idfilter:
            query = query.where(Bubble.id << idfilter)
        return query

    @classmethod
    def random_bubbles(cls):
        if not Bubble._buffer_bubbles or time.time() > Bubble._buffer_resync_time:
            query = Bubble.select()
            query = query.where(Bubble.listed == True)
            query = query.where(Bubble.password >> None)
            query = query.where(Bubble.type != Bubble.ENTERPRISE_BUBBLE)
            if random.randint(0, 1):
                query = query.order_by(Bubble.random.asc())
            else:
                query = query.order_by(Bubble.random.desc())
            query = query.limit(Bubble._buffer_size)
            Bubble._buffer_bubbles = tuple([bubble.id for bubble in query])
            Bubble._buffer_resync_time = time.time() + Bubble._buffer_time
        idfilter = set()
        if len(Bubble._buffer_bubbles) <= settings.GIVE_ME_BUBBLES_MAX:
            idfilter = Bubble._buffer_bubbles
        else:
            while len(idfilter) < settings.GIVE_ME_BUBBLES_MAX:
                idfilter.add(random.choice(Bubble._buffer_bubbles))
        query = Bubble.select().limit(settings.GIVE_ME_BUBBLES_MAX)
        if idfilter:
            query = query.where(Bubble.id << idfilter)
        return query

    def to_dict(self):
        return {"has_password": self.has_password,
                 "name": self.name,
                 "listed": self.listed,
                 "tags": self.tags,
                 "type": self.type,
                 "email": self.user.email,
                 "fname": self.user.fname,
                 "lname": self.user.lname,
                 "desc": self.desc}

    def set_password(self, password):
        if password is not None:
            self.password = security.generate_password_hash(password)
        else:
            self.password = None

    def check_password(self, password):
        if self.password is None:
            return password is None
        elif self.password is not None and password is not None:
            return security.check_password_hash(self.password, password)
        return False

    def save(self, force_insert=False):
        for ab in Bubble.select().where(Bubble.user == self.user, Bubble.name == self.name):
            if ab.id != self.id:
                raise ValueError("The user has 2 bubbles with same name")
        self.random = hashlib.sha256(str(random.random())).hexdigest()
        if not self.fsname:
            self.fsname = "bubble_{}".format(core.unique())
            os.makedirs(self.path)
        super(Bubble, self).save(force_insert)

    def delete_instance(self, recursive=False):
        if os.path.exists(self.path):
            shutil.rmtree(self.path)
        for img in Image.select().where(Image.bubble==self):
            img.delete_instance(True)
        super(Bubble, self).delete_instance(recursive)

    def set_tags(self, tags):
        for bxt in BubbleXTag.select().where(BubbleXTag.bubble == self):
            bxt.delete_instance(False)
        tags = set(tagname.strip().lower() for tagname in tags)
        for tagname in tags:
            tag = Tag.get_or_create(tagname=tagname)
            bxt = BubbleXTag()
            bxt.bubble = self
            bxt.tag = tag
            bxt.save()

    @property
    def path(self):
        return os.path.join(self.user.path, self.fsname)

    @property
    def tags(self):
        return [btx.tag.tagname for btx in self.bubblesxtags]

    @property
    def has_password(self):
        return self.password != None


#===============================================================================
# IMAGE TABLE
#===============================================================================

@autocreatetable
class Image(SocialFrameModel):

    STATUS_APPROVED = "approved"
    STATUS_REPORTED = "reported"
    STATUS_PENDING = "pending"

    ORIGINAL_SIZE = "original"

    bubble = db.ForeignKeyField(Bubble)
    desc = db.TextField(null=True)
    name = db.CharField(null=True)
    extension = db.CharField(max_length=10, null=True)
    uploaded_by = db.ForeignKeyField(User)
    status = db.CharField(choices=[(STATUS_APPROVED, STATUS_APPROVED),
                                   (STATUS_REPORTED, STATUS_REPORTED),
                                   (STATUS_PENDING, STATUS_PENDING)])
    def __unicode__(self):
        return self.path

    def to_dict(self):
        return {"sha": self.sha,
                 "filename": self.name,
                 "desc": self.desc,
                 "status": self.status,
                 "bubble": self.bubble.name,
                 "uploaded_by": self.uploaded_by.email}

    def regenerate(self):
        old_path = self.path if self.name else None
        seed = (str(core.unique()) + self.bubble.path + str(random.random()))
        self.name = hashlib.sha256(seed).hexdigest() + "." + self.extension
        if os.path.exists(self.path):
            self.name = uploadset.resolve_conflict(self.bubble.path, self.name)
        if not old_path:
            os.makedirs(self.path)
        else:
            shutil.move(old_path, self.path)
            self.save()

    def store(self, storage):
        if self.name:
            msg = "Image already stored"
            raise ValueError(msg)

        self.extension = os.path.splitext(storage.filename)[1][1:].lower()

        uploadset = uploads.UploadSet('uploadset', extensions=core.VALID_IMAGES)

        self.regenerate()
        uploadset._config = uploads.UploadConfiguration(self.path,
                                                        allow=core.VALID_IMAGES)
        imgsize = ImageSize()
        imgsize.image = self
        imgsize.filename = uploadset.save(storage, name=self.ORIGINAL_SIZE + ".")
        imgsize.size = self.ORIGINAL_SIZE
        self.save()
        imgsize.save()
        self.save()

    def get_size(self, size):
        if size not in settings.IMAGE_SIZES :
            msg = "'size' must be one of '{}', found: '{}'"
            raise ValueError(msg.format(settings.IMAGE_SIZES, size))
        for imgsize in ImageSize.select().where(ImageSize.image == self,
                                          ImageSize.size == size):
            return imgsize
        orig = ImageSize.get(ImageSize.image == self,
                             ImageSize.size == self.ORIGINAL_SIZE)
        imgsize = ImageSize()
        imgsize.image = self
        imgsize.filename = fileutils.image_resize(orig.path, self.extension,
                                                  size, orig.width, orig.height,
                                                  settings.IMAGE_SIZES[size])
        imgsize.size = size
        imgsize.save()
        return imgsize

    def delete_instance(self, recursive=False):
        if not os.path.exists(self.path):
            shutil.rmtree(self.path)
        for img in ImageSize.select().where(ImageSize.image==self):
            img.delete_instance(True)
        super(Image, self).delete_instance(recursive)

    @property
    def sha(self):
        return ImageSize.get(ImageSize.image == self,
                              ImageSize.size == self.ORIGINAL_SIZE).sha

    @property
    def path(self):
        return os.path.join(self.bubble.path, self.name)


#===============================================================================
# IMAGE SIZE TABLE
#===============================================================================

@autocreatetable
class ImageSize(SocialFrameModel):

    image = db.ForeignKeyField(Image, related_name="sizes")
    filename = db.CharField() #
    size = db.CharField() #
    width = db.IntegerField() #a
    height = db.IntegerField() #a
    sha = db.CharField() #a
    metadata = db.BooleanField(default=False) #a

    def save(self, force_insert=False):
        if not self.metadata:
            self.sha = fileutils.sha(self.path)
            self.width, self.height = fileutils.image_size(self.path)
            self.metadata = True
        super(ImageSize, self).save(force_insert)

    @property
    def path(self):
        return os.path.join(self.image.path, self.filename)


#===============================================================================
# TAG TABLE
#===============================================================================

@autocreatetable
class Tag(SocialFrameModel):

    tagname = db.CharField(unique=True)

    def save(self, force_insert=False):
        self.tagname = self.tagname.strip().lower()
        super(Tag, self).save(force_insert)


#===============================================================================
# BUBBLE X TAG Table
#===============================================================================

@autocreatetable
class BubbleXTag(SocialFrameModel):
    bubble = db.ForeignKeyField(Bubble, related_name="bubblesxtags")
    tag = db.ForeignKeyField(Tag)


#===============================================================================
# INVITATION
#===============================================================================

@autocreatetable
class BubbleInvitation(SocialFrameModel):

    sended_by = db.ForeignKeyField(User, null=True)
    to_user = db.ForeignKeyField(User, null=True)
    bubble = db.ForeignKeyField(Bubble)
    sended = db.BooleanField(default=False)
    email = db.CharField()
    invitation_hash = db.CharField()
    creation_datetime = db.DateTimeField()

    def save(self, force_insert=False):
        if not self.creation_datetime:
            self.creation_datetime = datetime.datetime.now()
        if not self.invitation_hash:
            self.invitation_hash = hashlib.sha256(str(random.random())).hexdigest()
        super(BubbleInvitation, self).save(force_insert)

    def validate_hash(self, invitation_hash):
        if self.invitation_hash:
            now = datetime.datetime.now()
            expiration = settings.HASH_EXPIRATION
            delta = datetime.timedelta(seconds=expiration)
            return now < self.creation_datetime + delta \
                  and invitation_hash == self.invitation_hash
        return False


#===============================================================================
# BUBBLE COMMITERS TABLE
#===============================================================================

@autocreatetable
class BubbleCommiter(SocialFrameModel):

    user = db.ForeignKeyField(User)
    bubble = db.ForeignKeyField(Bubble)
    _joined_datetime = db.DateTimeField(db_column="joined_datetime")

    def save(self, force_insert=False):
        if not self._joined_datetime:
            self._joined_datetime = datetime.datetime.now()
        for ci in BubbleCommiter.select().where(BubbleCommiter.user == self.user,
                                                 BubbleCommiter.bubble == self.bubble):
           if  ci.id != self.id:
                raise ValueError("This commiter already exists")
        super(BubbleCommiter, self).save(force_insert)

    @property
    def joined_datetime(self):
        return self._joined_datetime


#===============================================================================
# AUTHENTIFICATION
#===============================================================================

class CustomAuth(auth.Auth):
    def get_user_model(self):
        return User
    def get_urls(self):
        return ()

setattr(core, "APP_AUTH", CustomAuth(core.APP, core.APP_DB))


#===============================================================================
# MAIN
#===============================================================================

if __name__ == "__main__":
    print(__doc__)
